#pragma once
#include <PSX/GPU/gpu_auto_load_font.hpp>

namespace JabyEngine {
    namespace GPU {
        namespace SJIS {
            struct FontBuffer {                
                // A line of 16 Pixel
                struct Line {
                    Color pixel[BIOS_Font::Size.width];

                    void load(uint16_t bit_pattern) {
                        size_t px_idx = 0;
                        for(uint16_t shift = 1 << 15; shift; shift = shift >> 1) {
                            this->pixel[px_idx++] = bit_pattern & shift ? Color::from_rgb(0xFF, 0xFF, 0xFF) : Color::from_rgb(0x0, 0x0, 0x0);
                        }
                    }

                    static PositionU16 vram_offset(uint16_t tile_id) {
                        return tile_id_for16<PositionU16>(tile_id);
                    }
                };

                // Letters are actually only 16x15 but should be treated 16x16
                struct Letter : public GPU::internal::LinkedElementCreator<Letter> {
                    CPU2VRAM cmd;
                    Line     lines[BIOS_Font::Size.height - 1];
                    Line     empty_line;

                    void setup(const SizeU16& size) {
                        this->cmd.cmd  = GPU_IO::Command::CPU2VRAM_Blitting();
                        this->cmd.size = GPU_IO::Command::WidthHeight(size);
                        this->empty_line.load(0);
                    }

                    void load_to(const PositionU16& pos, const uint16_t* bit_map) {
                        this->cmd.pos = GPU_IO::Command::TopLeftPosition(pos);
                        for(auto& cur_line : this->lines) {
                            cur_line.load(__builtin_bswap16(*bit_map++));
                        }
                    }
                };

                //                            v double buffer do not change size without adjusting
                Letter::Linked  letter_buffer[2*4];
                uint8_t         free_idx;

                void setup() {
                    for(auto& letter : this->letter_buffer) {
                        letter->setup(SizeU16::create(16, 16));
                        letter.set_link_identitiy();
                    }
                    this->free_idx = 0;
                }

                void load_to(const PositionU16& pos, const uint16_t* bit_map) {
                    auto& cur_letter = this->letter_buffer[this->free_idx++];
                    
                    cur_letter->load_to(pos, bit_map);
                    if((this->free_idx&0x3) == 0) {
                        cur_letter.terminate();
                        GPU::render(this->letter_buffer[this->free_idx - 4]);
                        this->free_idx &= ~0x7;
                    }

                    else {
                        cur_letter.concat(this->letter_buffer[this->free_idx]);
                    }
                }

                void flush() {
                    if((this->free_idx&0x3) != 0) {
                        this->letter_buffer[this->free_idx - 1].terminate();

                        const auto idx = this->free_idx > 3 ? 4 : 0;
                        GPU::render(this->letter_buffer[idx]);
                    }
                }

                static PositionU16 vram_offset(uint16_t tile_id) {
                    return Line::vram_offset(tile_id);
                }             
            };

            struct SpecialChar {
                uint8_t base_offset;
                uint8_t tile_id;
            };

            struct RangeChar {
                SpecialChar start_char;
                uint8_t     length;
            };

            static_assert(BIOS_Font::Size == SizeU16::create(16, 16));
        }
    }
}